// Copyright 2024 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

package fips140

import (
	"internal/godebug"
	"syscall"
)

var fips140GODEBUG = godebug.New("#fips140")

// Enabled reports whether FIPS 140 mode is enabled by using GODEBUG, GOFIPS, GOLANG_FIPS,
// the 'requirefips' build tag, or any other platform-specific mechanism.
func Enabled() bool {
	return enabled
}

var enabled bool

// Disabled reports whether FIPS 140 mode is disabled by using GOFIPS or GOLANG_FIPS.
func Disabled() bool {
	return disabled
}

var disabled bool

// Message is a human-readable message about how [Enabled] was set.
var Message string

func init() {
	// TODO: Decide which environment variable to use.
	// See https://github.com/microsoft/go/issues/397.
	var ok bool
	value := fips140GODEBUG.Value()
	if value == "on" || value == "only" || value == "debug" {
		Message = "environment variable GODEBUG=fips140=" + value
		value = "1"
	} else if value, ok = syscall.Getenv("GOFIPS"); ok {
		Message = "environment variable GOFIPS"
	} else if value, ok = syscall.Getenv("GOLANG_FIPS"); ok {
		Message = "environment variable GOLANG_FIPS"
	} else if systemFIPSMode() {
		Message = "system FIPS mode"
		value = "1"
	}
	if value == "1" {
		enabled = true
	} else if value == "0" {
		disabled = true
	}
	if isRequireFIPS {
		if disabled {
			panic("the 'requirefips' build tag is enabled, but it conflicts " +
				"with the " + Message + "=" + value +
				" which would disable FIPS mode")
		}
		Message = "requirefips tag set"
		enabled = true
		return
	}
}
